@file:Suppress("ProblematicWhitespace")

package io.data2viz.charts.axis

import io.data2viz.charts.chart.Chart
import io.data2viz.charts.chart.mark.Scaler
import io.data2viz.charts.config.AxisConfig
import io.data2viz.charts.layout.ComponentSize
import io.data2viz.charts.layout.DrawingZone
import io.data2viz.geom.Size
import io.data2viz.charts.axis.internal.DefaultAxisRenderer as DAR


/**
 * Build a default axis visual component, can manage discrete or continuous values.
 *
 * The [DefaultAxisRenderer] is a simple axis component.
 * It can use custom formatter or pick default ones for numeric (Double), temporal (Date) or discrete values.
 */
public fun <VALUE> DefaultAxisRenderer(
    scaler: Scaler<VALUE>,
    title: String?,
    fmt: VALUE.() -> String,
    config: AxisConfig
): AxisRenderer = DAR(scaler, title, fmt, config)

/**
 * Axis visual component.
 */
public interface AxisRenderer {

    /**
     * Compute and returns the preferred [ComponentSize] based on a given available [Size].
     * This is called only before the first display or when the chart is resized and it is used to determine the
     * expected size of the [AxisRenderer].
     *
     * When given a "total size", the [AxisRenderer] is expected to:
     *  - take the full width (resp. height) as [ComponentSize.innerSize] if the axis is horizontal (resp. vertical)
     *  - compute its height (resp. width) as [ComponentSize.innerSize] if the axis is horizontal (resp. vertical)
     *  - indicates if it needs "padding" as [ComponentSize.padding], for example if some labels overflow
     *
     * @param totalSize: the default size given by the [Chart] layout, the axis must cover all the given size, for
     * example, a vertical axis innerSize.height is expected to be the same as the givenSize.height
     *
     * @return [ComponentSize]: the minimum axis size (as [ComponentSize.innerSize]) and the "overflow"
     * of the [AxisRenderer] as [ComponentSize.padding].
     **/
    public fun getSize(totalSize: Size) : ComponentSize

    /**
     * This function is called every refresh of the AxisRenderer (first display, zoom, pan, resize....).
     * If the AxisRenderer needs to compute some label rotation you can compute it here.
     *
     * @see getSize
     */
    public fun computeLabelSpaceAndRotation()

    @Deprecated("Use drawBackgroundElements() instead", ReplaceWith("drawBackgroundElements(drawingZone)"))
    public fun drawBackground(drawingZone: DrawingZone): Unit = drawBackgroundElements(drawingZone)

    /**
     * Draw or redraw the background elements associated to the axis : gridLines & alternateBackgrounds
     *
     * @param drawingZone: this is the chart's main drawing zone, the [AxisRenderer] is expected to draw alternate
     * backgrounds and gridlines directly on this [DrawingZone].
     */
    public fun drawBackgroundElements(drawingZone: DrawingZone)

    @Deprecated("Use drawAxis() instead", ReplaceWith("drawAxis(drawingZone)"))
    public fun drawForeground(drawingZone: DrawingZone): Unit = drawAxis(drawingZone)

    /**
     * Draw or redraw the axis, it will appear in front of the marks.
     *
     * @param drawingZone: this is the [AxisRenderer] drawing zone, it draws directly on this zone, every coordinates
     * are relative to this [DrawingZone].
     */
    public fun drawAxis(drawingZone: DrawingZone)

}
