package io.data2viz.charts.chart

import io.data2viz.charts.chart.internal.ChartImpl
import io.data2viz.charts.chart.mark.*
import io.data2viz.charts.config.ChartConfig
import io.data2viz.charts.config.CursorConfig
import io.data2viz.charts.config.configuration
import io.data2viz.charts.core.*
import io.data2viz.charts.core.Dataset
import io.data2viz.charts.core.Padding
import io.data2viz.charts.core.Zoom
import io.data2viz.charts.dimension.Discrete
import io.data2viz.charts.dimension.DiscreteDimension
import io.data2viz.charts.event.*
import io.data2viz.charts.layout.*
import io.data2viz.charts.viz.VizContainer
import io.data2viz.color.Color
import io.data2viz.geom.Size
import io.data2viz.viz.HasMetaKeys


/**
 * MarkCreator is simply a name for a lambda that instantiates a [Mark] when called.
 *
 * @returns a [Mark]
 */
internal typealias MarkCreator<DOMAIN> = (index: Int) -> Mark<DOMAIN>

/**
 * Builds a [Chart] using this [VizContainer].
 *
 * @param data The original collection of DOMAIN objects.
 * @param config A configuration instance that will be copied for the current chart.
 * @param init The initialization lambda that will be applied to the [Chart], this is where all the magic happens!
 */
public fun <DOMAIN> VizContainer.chart(
    data: List<DOMAIN>,
    config: ChartConfig = configuration {  },
    init: Chart<DOMAIN>.() -> Unit,
): Chart<DOMAIN> =
    ChartImpl(this, data, config.deepCopy(), init)


/**
 * The [Chart] class is the main entry point when creating a chart,
 * use the [VizContainer.chart] function to create one.
 */
public interface Chart<DOMAIN>:
    EventListener<DOMAIN>,
    ChartController<DOMAIN>
{
    /**
     * The external size of the chart. Initialized with the viz container size.
     *
     * On write value, it modifies the underlying platform
     * element of the VizContainer (HTML Div, Jfx Pane, ...).
     */
    @Deprecated("You must not use this value to set the size of a Chart. The size should be set directly on the VizContainer")
    public var size: Size

    /**
     * The main configuration object [ChartConfig], stores all configurations for the underlying marks, axes, ...
     * The [ChartConfig] holds information about:
     *  - The styling of this [Chart]: fonts, colors, padding..
     *  - The interactivity of this [Chart]: zoom, selection, cursor...
     *  - etc...
     *
     *  You can build your own [ChartConfig] or uses some default ones:
     *   - [ChartConfig.default]
     *   - [ChartConfig.light]
     *   - ...
     */
    public val config: ChartConfig

    /**
     * Allow to override the default config or the config passed into the Chart.
     */
    public fun config(init: ChartConfig.()-> Unit)

    /**
     * The chart [Dataset], a data container giving access to all the original data and several useful properties.
     */
    public val dataset: Dataset<DOMAIN>

    /**
     * Update the current dataset of the chart, only if it is a new [List] (compared by reference).
     */
    public fun update(newData: List<DOMAIN>)

    /**
     * The series [Discrete] dimension is used to "split" your data into multiple series allowing to display
     * them differently.
     *
     * By default there is only 1 series, and all data is part of this series.
     * When specifying a series [Dimension], the data will be split into several [List], each series will be available
     * through [Dataset.dataBySeries] and [Mark] compatible with multiple series will handle them accordingly.
     *
     * For example, when defining multiple series a [LineMark] will draw a line for each series.
     * Having multiple series will allow stacking of compatible [Mark] like [ColumnMark]...
     */
    public var series: DiscreteDimension<DOMAIN, *>

    /**
     * The current [Zoom] for this [Chart].
     * This is an absolute zoom value, combining all the modifications (pan, zoom-in & zoom-out...) that have been
     * applied to the chart.
     */
    public val zoom: Zoom

    /**
     * Force an absolute zoom value, see [zoom].
     * This can be useful if you want to save/load zoom values on a chart.
     */
    @ExperimentalCharts
    public fun applyZoom(zoom: Zoom)

    /**
     * Add a [Mark] to the current chart.
     * It should only be called if you implement your own custom Mark.
     *
     * @param markCreator the lambda that instantiates the [Mark] when called.
     */
    public fun addMark(markCreator: MarkCreator<DOMAIN>)

    /**
     * The [Tooltip] initialization DSL.
     *
     * @param init the [Tooltip] initialization lambda, use this DSL to setup properties of the [Tooltip]
     */
    public fun tooltip(init: Tooltip<DOMAIN>.() -> Unit)

    /**
     * The [Legend] initialization DSL.
     *
     * @param init the [Legend] initialization lambda, use this DSL to setup properties of the [Legend]
     */
    public fun legend(init: Legend<DOMAIN>.() -> Unit)

    /**
     * Give a title to your Chart.
     */
    public var title: String?

    /**
     * The [Size] of the charting zone (ie. except title, legends, axes...)
     */
    public val chartingSize: Size

    //Deprecates

    @Deprecated("Use config.padding", replaceWith = ReplaceWith("config.padding"))
    public var padding: Padding

    @Deprecated("Use config.backgroundColor", replaceWith = ReplaceWith("config.backgroundColor"))
    public var backgroundColor: Color

    @Deprecated("Use config.performanceMode", replaceWith = ReplaceWith("config.performanceMode"))
    public var performanceMode: Boolean

    @Deprecated("Use config.events.highlightMode", replaceWith = ReplaceWith("config.events.highlightMode"))
    public var highlightMode: HighlightMode

    @Deprecated("Use config.events.triggerMode", replaceWith = ReplaceWith("config.events.triggerMode"))
    public var triggerMode: TriggerMode

    @Deprecated("Use config.events.selectionMode", replaceWith = ReplaceWith("config.events.selectionMode"))
    public var selectionMode: SelectionMode

    @Deprecated("Use config.events.zoomMode", replaceWith = ReplaceWith("config.events.zoomMode"))
    public var zoomMode: ZoomMode

    @Deprecated("Use config.events.zoomEnableOnMouseWheel", replaceWith = ReplaceWith("config.events.zoomEnableOnMouseWheel"))
    public var zoomEnableOnMouseWheel: HasMetaKeys.() -> Boolean

    @Deprecated("Use config.events.panMode", replaceWith = ReplaceWith("config.events.panMode"))
    public var panMode: PanMode

    @Deprecated("Use config.cursor", replaceWith = ReplaceWith("config.cursor"))
    public fun cursor(init: CursorConfig.() -> Unit): Unit = config.cursor(init)

}
