package io.data2viz.charts.chart.mark

import io.data2viz.charts.chart.Chart
import io.data2viz.charts.chart.mark.internal.AreaMarkImpl
import io.data2viz.charts.config.AxisConfig
import io.data2viz.charts.config.EventConfig
import io.data2viz.charts.config.MarkConfig
import io.data2viz.charts.core.Dataset
import io.data2viz.charts.core.Datum
import io.data2viz.charts.dimension.Dimension
import io.data2viz.charts.dimension.Discrete
import io.data2viz.charts.dimension.Quantitative

/**
 * Instantiate a new [AreaMark].
 *
 * See [Area chart](https://en.wikipedia.org/wiki/Area_chart) for more info.
 *
 * @param dataset Your chart's [Dataset]
 * @param x The [Quantitative] dimension for the X axis
 * @param y The [Discrete] dimension for the Y axis
 * @param init The initialization lambda for this [Mark]
 * @param config The [MarkConfig] to configure this [Mark]
 * @param xAxisConfig The [AxisConfig] to configure this mark's X axis
 * @param yAxisConfig The [AxisConfig] to configure this mark's Y axis
 */
public fun <DOMAIN, YVAL> areaMark(
    markIndex: Int,
    dataset: Dataset<DOMAIN>,
    x: Quantitative<DOMAIN>,
    y: Discrete<DOMAIN, YVAL>,
    init: AreaMark<DOMAIN, Double?, YVAL>.() -> Unit,
    eventConfig: EventConfig,
    config: MarkConfig,
    xAxisConfig: AxisConfig,
    yAxisConfig: AxisConfig
): AreaMark<DOMAIN, Double?, YVAL> = AreaMarkImpl(markIndex, dataset, x, y, false, init, eventConfig, config, xAxisConfig, yAxisConfig)

/**
 * Instantiate a new [AreaMark].
 *
 * See [Area chart](https://en.wikipedia.org/wiki/Area_chart) for more info.
 *
 * @param dataset Your chart's [Dataset]
 * @param x The [Discrete] dimension for the X axis
 * @param y The [Quantitative] dimension for the Y axis
 * @param init The initialization lambda for this [Mark]
 * @param config The [MarkConfig] to configure this [Mark]
 * @param xAxisConfig The [AxisConfig] to configure this mark's X axis
 * @param yAxisConfig The [AxisConfig] to configure this mark's Y axis
 */
public fun <DOMAIN, XVAL> areaMark(
    markIndex: Int,
    dataset: Dataset<DOMAIN>,
    x: Discrete<DOMAIN, XVAL>,
    y: Quantitative<DOMAIN>,
    init: AreaMark<DOMAIN, XVAL, Double?>.() -> Unit,
    eventConfig: EventConfig,
    config: MarkConfig,
    xAxisConfig: AxisConfig,
    yAxisConfig: AxisConfig
): AreaMark<DOMAIN, XVAL, Double?> = AreaMarkImpl(markIndex, dataset, x, y, true, init, eventConfig, config, xAxisConfig, yAxisConfig)

/**
 * Add an [AreaMark] to the current [Chart].
 *
 * See [Area chart](https://en.wikipedia.org/wiki/Area_chart) for more info.
 *
 * @param x The [Quantitative] dimension for the X axis
 * @param y The [Discrete] dimension for the Y axis
 * @param init The initialization lambda for this [Mark]
 */
public fun <DOMAIN, YVAL,> Chart<DOMAIN>.area(
    x: Quantitative<DOMAIN>,
    y: Discrete<DOMAIN, YVAL>,
    init: AreaMark<DOMAIN, Double?, YVAL>.() -> Unit = {}
): Unit = addMark { markIndex -> areaMark(markIndex, dataset, x, y, init, config.events, config.mark, config.x, config.y) }

/**
 * Add an [AreaMark] to the current [Chart].
 *
 * See [Area chart](https://en.wikipedia.org/wiki/Area_chart) for more info.
 *
 * @param x The [Discrete] dimension for the X axis
 * @param y The [Quantitative] dimension for the Y axis
 * @param init The initialization lambda for this [Mark]
 */
public fun <DOMAIN, XVAL,> Chart<DOMAIN>.area(
    x: Discrete<DOMAIN, XVAL>,
    y: Quantitative<DOMAIN>,
    init: AreaMark<DOMAIN, XVAL, Double?>.() -> Unit = {}
): Unit = addMark { markIndex -> areaMark(markIndex, dataset, x, y, init, config.events, config.mark, config.x, config.y) }

/**
 * Draw an area for a given [Dataset] and X and Y [Dimension]s.
 */
public interface AreaMark<DOMAIN, XVAL, YVAL>:
    HasFill<DOMAIN>,
    HasSize<DOMAIN>,
    HasMarker<DOMAIN>,
    HasCurve,
    HasStack,
    HasStrokeColor<DOMAIN>,
    HasStrokeWidth<DOMAIN>,
    HasStrokeLine<DOMAIN>,
    HasMarkDecorator<DOMAIN>,
    HasHighlightDecorator<DOMAIN>,
    HasSelectionDecorator<DOMAIN>,
    HasEventZones<DOMAIN>,
    HasLegendDecorator<DOMAIN>,
    HasCustomTooltip<DOMAIN>,
    HasHighlightMode,
    HasSelectionMode,
    MarkXY<DOMAIN, XVAL, YVAL> {

    // TODO HasBaseline ?
    /**
     * The "base line" [Double] value of each area, default to the constant value 0.
     * Use a variable [baseline] value to display "ranges".
     */
    public var baseline: Datum<DOMAIN>.() -> Double
}

@Deprecated(
    "Replace with AreaMark.",
    ReplaceWith("AreaMark<DOMAIN, XVAL, YVAL>", "io.data2viz.charts.chart.mark.AreaMark")
)
public typealias Area<DOMAIN, XVAL, YVAL> = AreaMark<DOMAIN, XVAL, YVAL>
