@file:Suppress("UNCHECKED_CAST")

package io.data2viz.charts.chart.mark

import io.data2viz.charts.axis.AxisRenderer
import io.data2viz.charts.axis.internal.DefaultAxisRenderer
import io.data2viz.charts.chart.Chart
import io.data2viz.charts.config.AxisConfig
import io.data2viz.charts.core.*
import io.data2viz.charts.dimension.*
import io.data2viz.charts.internal.ResettableLazy
import io.data2viz.charts.layout.DrawingZone
import io.data2viz.charts.layout.internal.CElement
import io.data2viz.charts.logging.KLogging
import io.data2viz.math.Percent
import io.data2viz.math.pct
import io.data2viz.scale.*
import io.data2viz.timeFormat.Locales
import kotlinx.datetime.Instant
import kotlinx.datetime.Instant.Companion.fromEpochMilliseconds
import kotlinx.datetime.LocalDateTime
import kotlinx.datetime.TimeZone
import kotlinx.datetime.toInstant
import kotlin.math.abs
import kotlin.math.max
import kotlin.math.min
import kotlin.math.round

private val MINIMUM_TEMPORAL = LocalDateTime(1000, 1, 1, 0, 0).toInstant(TimeZone.UTC)
private val MAXIMUM_TEMPORAL = LocalDateTime(3000, 1, 1, 0, 0).toInstant(TimeZone.UTC)

private val logger = KLogging.logger("io.data2viz.charts.chart")

// TODO : when an axis.isChild == true, there is no need for a scale.

/**
 * The core "axis" component for a given [Mark].
 *
 * Generally, a [Chart] will have several [Dimension]s that can be shared between the different [Mark]s, the [Axis]
 * is the concrete use of a [Dimension] for a given [Mark].
 *
 * @see [AxisRenderer] The visual part of the axis component.
 */
public open class Axis<DOMAIN, VALUE>(
    private var data: List<Datum<DOMAIN>>,

    /**
     * The [Dimension] this [Axis] is build from.
     */
    public val dimension: Dimension<DOMAIN, VALUE>,

    /**
     * The [AxisConfig] of this Axis
     */
    public val config: AxisConfig,

    /**
     * Does this [Axis] requires a bandwidth or not?
     * When creating an axis for a bar chart for example, you require a bandwidth, for a line chart your don't.
     */
    defaultUseBandwidth: Boolean
)

    : Scaler<VALUE>,
    AxisConfig by config.also( { dimension.axisConfigInitializer?.invoke(config) } ) {

    /**
     * You can override the "bandwidth" attribute of a discrete axis.
     * For example if you want to synchronize 2 marks along a shared discrete X-axis, a bar mark (which use bandwidth)
     * and a line mark (which does not), the best option is to enable the bandwidth of the x-axis of the line mark,
     * mark.x.useBandwidth = true.
     */
    public var useBandwidth: Boolean = defaultUseBandwidth
        set(value) {
            field = value
            _scale = defaultScale()
            _tickValues = null
            bandwidthRatio = bandwidthRatio
            _zoomScale = _scale
        }

    @Deprecated("Use dimension instead.", ReplaceWith("dimension"))
    public val dim: Dimension<DOMAIN, VALUE> = dimension

    // TODO check if really needed ? or must pass through getPosition()
    /**
     * A [List] of VALUE corresponding to every DOMAIN element pass through this [Axis]'s [Dimension] accessor.
     */
    private val valuesLazy = ResettableLazy {
        logger.verbose("Initialize values for $this")
        data.map { dimension.accessor(it) }
    }
    public val values: List<VALUE> by valuesLazy

    internal fun updateDataset(newDataset: Dataset<DOMAIN>) {
        data = newDataset.data
        _tickValues = null
        valuesLazy.invalidate()
    }

    /**
     * Changing the bandwidth must launch a rescale
     */
    override var bandwidthRatio: Percent
        get() = config.bandwidthRatio
        set(value) {
            config.bandwidthRatio = value
            if (_scale is BandScale) {
                val pi = if (useBandwidth) (100.pct - config.bandwidthRatio) else 100.pct
                val po = if (useBandwidth) (pi / 2.0) else 0.pct
                (_scale as BandScale).paddingInner = pi
                (_scale as BandScale).paddingOuter = po
                (_zoomScale as BandScale).paddingInner = pi
                (_zoomScale as BandScale).paddingOuter = po
            }
        }

//    /**
//     * Changing the "pointAxis" must launch a rescale
//     */
//    override var enableBandwidth: Bandwidth
//        get() = config.enableBandwidth
//        set(value) {
//            config.enableBandwidth = value
//            if (_scale is BandScale) {
//                val pi = if (config.enableBandwidth.isEnabled) (100.pct - config.bandwidthRatio) else 100.pct
//                val po = if (config.enableBandwidth.isEnabled) (pi / 2.0) else 0.pct
//                (_scale as BandScale).paddingInner = pi
//                (_scale as BandScale).paddingOuter = po
//                (_zoomScale as BandScale).paddingInner = pi
//                (_zoomScale as BandScale).paddingOuter = po
//            }
//        }



    //Scaler implementation: TODO to remove
    private var _tickValues: List<VALUE>? = null
        set(value) {
            field = value
            updateTickPrecision()
        }

    /**
     * The "max" of the axis is the maximum numeric amount to be displayed
     * This is used when trying to find a suitable formatter.
     */
    private var _max: Double = 1.0

    /**
     * The "spread" of the axis is the (max value - min value)
     * This is used to compute the _step.
     */
    private var _spread: Double = 1.0

    /**
     * The "step" of the axis is the minimum numeric step between 2 ticks.
     * This is used when trying to find a suitable formatter.
     */
    private var _step: Double = 1.0

    override val tickValues: List<VALUE>
        get() {
            if (_tickValues == null) {
                _tickValues = defaultTicksImpl()
            }
            return _tickValues!!
        }

    private fun updateTickPrecision() {
        if (!_tickValues.isNullOrEmpty() && _tickValues!!.filterNot { it == null }.firstOrNull() is Double) {
            _max = (tickValues as List<Double>).maxBy { abs(it) }
            _spread = computeDoubleRange()
            _step = _spread / max(1, tickValues.size - 1)
        }
    }

    override fun scaleValue(value: VALUE): Double =
        scale(value) + if (scale is BandedScale) ((scale as BandedScale).bandwidth / 2.0) else .0

    override fun invert(value: Double): VALUE = when (val sc = scale) {
        is InvertableScale<VALUE, Double> -> sc.invert(value)
        else -> error("Scale is not invertable :: $scale")
    }


    /**
     * Return the position of this [Datum] object, along this [Axis], or null if this object has no position.
     *
     * @param datum the [Datum] object we want to display on screen.
     * @return the position in [Double], relative to the current [DrawingZone] context.
     */
    public fun getPosition(datum: Datum<DOMAIN>): Double? =
        values[datum.indexInData]?.let { scaleValue(it) }

    /**
     * Return the labels for a [Datum] from this Axis [Dimension] point of view.
     */
    public fun getLabel(datum: Datum<DOMAIN>): String = getLabel(values[datum.indexInData])

    /**
     * Return the labels for a given [VALUE] from this Axis [Dimension] point of view.
     */
    public fun getLabel(domainValue: VALUE): String =
        "${(dimension.name?.plus(": ") ?: "")}${valueFormatter(domainValue)}"

    /**
     * In case of shared axis/dimension
     * (if the axis represents the same dimension of another mark or the axis use another axis )
     *
     * Note: the parent is always the last parent, or the "ancestor" as a child dimension cannot have its own child.
     */
    internal var parent: Axis<DOMAIN, VALUE>? = null

    internal fun setParentAxis(axis: Axis<DOMAIN, *>) {
        parent = axis as Axis<DOMAIN, VALUE> // Can't be done using the setter outside the class.
    }

    /**
     * Return the parent's dimension, or the current one if there is no parent.
     */
    internal val parentOrCurrent
        get() = parent ?: this

    /**
     * If it has a [parent]
     */
    internal val isChild: Boolean
        get() = parent != null

    /**
     * This is the formatter for the axis labels, it can be set from the dimension formatter or use
     * a "smartFormatter" that is tailored to display the axis labels at a "smart" precision.
     */
    private val axisFormatter by lazy { dimension.formatter ?: getSmartFormatterForAxis() }

    /**
     * This is the formatter for values, it can be set from the dimension formatter or use
     * a "smartFormatter" that is tailored to display the axis labels at a "smart" precision.
     */
    private val valueFormatter
        get() = dimension.formatter ?: getSmartFormatterForValues()

    private var _scale: Scale<VALUE, Double> = defaultScale()
        set(value) {
            field = value
            _tickValues = null
        }

    private var _zoomScale: Scale<VALUE, Double> = _scale
        set(value) {
            field = value
            _tickValues = null
        }

    private fun computeDoubleRange(): Double {
        val doubles = tickValues as List<Double>
        val max = doubles.maxOrNull() ?: 1.0
        val min = doubles.minOrNull() ?: .0
        return abs(max - min)
    }

    /**
     * The "Scale" is the interpolator that is used to transform one of your value (could be a Double, an Instant...)
     * to a screen position.
     * The [Scale] is created from your dimension, for example a [Quantitative] dimension will create a continuous
     * linear numeric scale by default, and a [Discrete] dimension will create a discrete scale.
     *
     * You can use your own scale if it is compatible, for example you can use a Logarithmic scale
     * ([Scales.Continuous.log]) for a [Quantitative] dimension.
     */
    public var scale: Scale<VALUE, Double>
        set(value) {
            require(isChild.not()) { "You cannot set a scale on an axis that uses a \"child\" dimension, set it on the axis that use the parent dimension." }
            _scale = value
            _zoomScale = value
        }
        get() = parent?.scale ?: _zoomScale


    /**
     * This axis renderer, use [DefaultAxisRenderer] by default but you can specify your own custom [AxisRenderer].
     */
    public var renderer: () -> AxisRenderer = {
        DefaultAxisRenderer(this, title ?: dimension.name, axisFormatter, this)
    }

    internal val _axisRenderer: AxisRenderer by lazy { renderer() }

    /**
     * The axis title, if no title is set, the axis uses the dimension's name.
     */
    public var title: String? = null

    /**
     * The minimal value your axis can display, constrain your zoom, pan... to this limit.
     * To avoid misconception, this can only apply to a "parent axis".
     *
     * Note: if you want to display your axis using a given range of values, use [start] / [end] instead.
     */
    public var min: VALUE? = null
        set(value) {
            // TODO should use isChild.not() but the parent is not initialized at this time !!
            require(dimension.isChildDimension.not()) { "You cannot set a \"min\" value to a child axis, set this value to the parent axis (the axis which is created from the parent dimension) instead." }
            field = value
        }

    /**
     * The maximal value your axis can display, constrain your zoom, pan... to this limit.
     * To avoid misconception, this can only apply to a "parent axis".
     *
     * Note: if you want to display your axis using a given range of values, use [start] / [end] instead.
     */
    public var max: VALUE? = null
        set(value) {
            // TODO should use isChild.not() but the parent is not initialized at this time !!
            require(dimension.isChildDimension.not()) { "You cannot set a \"max\" value to a child axis, set this value to the parent axis (the axis which is created from the parent dimension) instead." }
            field = value
        }

    /**
     * The starting value of your axis visual range.
     * To avoid misconception, this can only apply to a "parent axis".
     *
     * Note: If you want to constrain your axis to given limits, use [min] / [max] instead.
     */
    public var start: VALUE? = null
        set(value) {
            // TODO should use isChild.not() but the parent is not initialized at this time !!
            require(dimension.isChildDimension.not()) { "You cannot set a \"start\" value to a child axis, set this value to the parent axis (the axis which is created from the parent dimension) instead." }
            field = value
        }

    /**
     * The ending value of your axis visual range.
     * To avoid misconception, this can only apply to a "parent axis".
     *
     * Note: If you want to constrain your axis to given limits, use [min] / [max] instead.
     */
    public var end: VALUE? = null
        set(value) {
            // TODO should use isChild.not() but the parent is not initialized at this time !!
            require(dimension.isChildDimension.not()) { "You cannot set an \"end\" value to a child axis, set this value to the parent axis (the axis which is created from the parent dimension) instead." }
            field = value
        }

    // Scales -----

    internal fun defaultScale(): Scale<VALUE, Double> = when (dimension) {
        is Quantitative<*> -> Scales.Continuous.linear {
            domain = listOf(Double.POSITIVE_INFINITY, Double.NEGATIVE_INFINITY)
        } as Scale<VALUE, Double>
        is Temporal<*> -> Scales.Continuous.time {
            domain = listOf(
                LocalDateTime(10000, 1, 1, 0, 0).toInstant(TimeZone.UTC),
                LocalDateTime(-10000, 1, 1, 0, 0).toInstant(TimeZone.UTC)
            )
        } as Scale<VALUE, Double>
        is Discrete<*, *> -> Scales.Discrete.band {
            val pi = if (useBandwidth) (100.pct - config.bandwidthRatio) else 100.pct
            val po = if (useBandwidth) (pi / 2.0) else 0.pct
            paddingInner = pi
            paddingOuter = po

            // barAlignment has been removed
//            align = barAlignment

            // DO NOT ACTIVATE ROUND! if there is a lot of data (ie 1000+ categories) you'll loose A LOT of space
            round = false
        }
        else -> error("Unknown dimension")
    }

    private var minMaxUpdated = false

    /**
     * Compute the (parent) scale domain from the dataset values.
     */
    internal fun updateScaleFromValues() {
//        if (minMaxUpdated) error("UpdateScaleFromValues should be called only once.")
        minMaxUpdated = true
        when (dimension) {
            is Quantitative<*> -> {
                val continuousScale = scale as ContinuousScale<Double, Double>
                val minValue = (values as List<Double?>).filterNotNull().minOrNull() ?: Double.POSITIVE_INFINITY
                val maxValue = (values as List<Double?>).filterNotNull().maxOrNull() ?: Double.NEGATIVE_INFINITY
                val newMin = min(minValue, continuousScale.domain.first())
                val newMax = max(maxValue, continuousScale.domain.last())
                continuousScale.domain = listOf(newMin, newMax)
            }
            is Temporal<*> -> {
                val continuousScale = scale as ContinuousScale<Instant, Double>
                val minValue = (values as List<Instant>).minOrNull() ?: MAXIMUM_TEMPORAL
                val maxValue = (values as List<Instant>).maxOrNull() ?: MINIMUM_TEMPORAL
                val newMin = listOfNotNull(minValue, continuousScale.domain.first()).minByOrNull { it }!!
                val newMax = listOfNotNull(maxValue, continuousScale.domain.last()).maxByOrNull { it }!!
                continuousScale.domain = listOf(newMin, newMax)
            }
            is Discrete<*, *>, is Constant -> {
                val dimensionValues: List<VALUE?> = values.toSet().toList()
                (scale as DiscreteDomain<VALUE?>).domain =
                    ((scale as DiscreteDomain<VALUE?>).domain + dimensionValues).distinct()
            }

        }
    }

    /**
     * Apply the start/end values to the domain
     */
    internal fun restrictScaleToStartEnd() {
        if (dimension.isChildDimension) return
        when (dimension) {
            is Quantitative<*> -> {
                val continuousScale = scale as ContinuousScale<Double, Double>
                val updatedMin = (start ?: continuousScale.domain.first()) as Double
                val updatedMax = (end ?: continuousScale.domain.last()) as Double
                continuousScale.domain = listOf(updatedMin, updatedMax)
            }
            is Temporal<*> -> {
                val continuousScale = scale as ContinuousScale<Instant, Double>
                val updatedMin = (start ?: continuousScale.domain.first()) as Instant
                val updatedMax = (end ?: continuousScale.domain.last()) as Instant
                continuousScale.domain = listOf(updatedMin, updatedMax)
            }
            else -> Unit
        }
    }

    /**
     * Nice the domain if needed
     */
    internal fun niceDomain(rangeSize: Double) {
        if (config.autoFit.not() ||
            dimension.isDiscrete ||
            start != null ||
            end != null ||
            dimension.isChildDimension ||
            dimension is Temporal
        ) return

        when (dimension) {
            is Quantitative<*> -> {
                val continuousScale = scale as NiceableScale<Double>
                val ticks = getDefaultContinuousTicks(scale as Tickable<*>, rangeSize).size
                continuousScale.nice(ticks)

                val domainMin = continuousScale.domain.first()
                val domainMax = continuousScale.domain.last()
                val finalMin = if (min != null) max((min as Double), domainMin) else domainMin
                val finalMax = if (max != null) min((max as Double), domainMax) else domainMax
                continuousScale.domain = listOf(finalMin, finalMax)
            }
            is Temporal<*> -> {
                val continuousScale = scale as NiceableScale<Instant>
                val ticks = getDefaultContinuousTicks(scale as Tickable<*>, rangeSize).size
                continuousScale.nice(ticks)

                val domainMin = continuousScale.domain.first()
                val domainMax = continuousScale.domain.last()
                val finalMin = if (min != null) listOf((min as Instant), domainMin).maxByOrNull { it }!! else domainMin
                val finalMax = if (max != null) listOf((max as Instant), domainMax).minByOrNull { it }!! else domainMax
                continuousScale.domain = listOf(finalMin, finalMax)
            }
            else -> Unit
        }
    }

    /**
     * Apply the min/max constraints to the domain
     */
    internal fun restrictScaleToMinMax() {
        if (dimension.isChildDimension) return
        when (dimension) {
            is Quantitative<*> -> {
                val continuousScale = scale as ContinuousScale<Double, Double>
                val domainMin = continuousScale.domain.first()
                val domainMax = continuousScale.domain.last()
                val finalMin = if (min != null) max((min as Double), domainMin) else domainMin
                val finalMax = if (max != null) min((max as Double), domainMax) else domainMax
                continuousScale.domain = listOf(finalMin, finalMax)
            }
            is Temporal<*> -> {
                val continuousScale = scale as ContinuousScale<Instant, Double>
                val domainMin = continuousScale.domain.first()
                val domainMax = continuousScale.domain.last()
                val finalMin = if (min != null) listOf((min as Instant), domainMin).maxByOrNull { it }!! else domainMin
                val finalMax = if (max != null) listOf((max as Instant), domainMax).minByOrNull { it }!! else domainMax
                continuousScale.domain = listOf(finalMin, finalMax)
            }
            else -> Unit
        }
    }

    internal fun updateScaleRange(from: Double, to: Double) {
        if (dimension.isChildDimension) return

        when (dimension) {
            is Quantitative<*> -> (_scale as ContinuousRangeScale<*, Double>).range = listOf(from, to)
            is Temporal<*> -> (_scale as ContinuousRangeScale<*, Double>).range = listOf(from, to)
            is Discrete<*, *> -> (_scale as StrictlyContinuousRange<*, Double>).range = StrictlyContinuous(from, to)
            else -> error("Unknown dimension $dimension")
        }
        logger.verbose("updateScaleRange:: ${_scale.toDescription()}")
    }

    // TODO doc and optimization (computeDoubleRange only once....)
    private fun getSmartFormatterForAxis(): VALUE.() -> String {
        return {
            when (this) {
                is Labelled -> label
                is Double? -> formatSmartDouble(io.data2viz.format.Locales.en_US, _step, _max)
                is Instant -> formatSmartInstant()
                else -> toString()
            }
        }
    }

    private fun Instant.formatSmartInstant() = TimeZone.UTC.(Locales.en_US.autoFormat())(this)


    private fun getSmartFormatterForValues(): VALUE.() -> String {
        return {
            when (this) {
                is Labelled -> label
                is Double? -> formatToDecimal()
                is Instant -> TimeZone.UTC.(Locales.en_US.autoFormat())(this)
                else -> toString()
            }
        }
    }

    /**
     * To avoid ratio rounding to things like 1.9999999998, round it to a "significant" value
     */
    private fun Double.roundToSignificantRatio(): Double = round(this * 1000.0) / 1000.0

    /**
     * Check of the zoom is valid, or return an updated zoom.
     * For example, if an [Axis] has some limit values and forbid zooming beyond them, it will return
     * an updated version of the zoom constrained by these limits.
     */
    internal fun validateZoom(zoom: DirectedZoom): DirectedZoom {
        if (dimension.isChildDimension) return zoom
        if (zoom == DirectedZoom.IDENTITY) return zoom

        _zoomScale = _scale
        val sc = scale
        return when (dimension) {
            is Quantitative -> {
                val zoomedScale = zoom.rescale(sc as ContinuousScale<Double, Double>) as Scale<VALUE, Double>

                val domain = (zoomedScale as ContinuousScale<Double, Double>).domain
                val newMin =
                    if (min != null && domain.first() < (min!! as Double)) (min!! as Double) else domain.first()
                val newMax = if (max != null && domain.last() > (max!! as Double)) (max!! as Double) else domain.last()

                var updatedZoom = zoom

                if (newMin > domain.first()) {
                    val diff = newMin - domain.first()
                    val newMaxMoved = newMax + diff
                    val updatedMax = if (max != null) min((max!! as Double), newMaxMoved) else newMaxMoved

                    val finalScale = (_scale.copy() as ContinuousScale<Double, Double>).also {
                        it.domain = listOf(newMin, updatedMax)
                    }
                    updatedZoom = finalScale.computeZoom((_scale as ContinuousScale<Double, Double>))
                } else if (newMax < domain.last()) {
                    val diff = domain.last() - newMax
                    val newMinMoved = newMin - diff
                    val updatedMin = if (min != null) max((min!! as Double), newMinMoved) else newMinMoved

                    val finalScale = (_scale.copy() as ContinuousScale<Double, Double>).also {
                        it.domain = listOf(updatedMin, newMax)
                    }
                    updatedZoom = finalScale.computeZoom((_scale as ContinuousScale<Double, Double>))
                }

                updatedZoom
            }
            is Temporal -> {
                val zoomedScale = zoom.rescale(sc as ContinuousScale<Instant, Double>) as Scale<VALUE, Double>

                val domain = (zoomedScale as ContinuousScale<Instant, Double>).domain
                val newMin =
                    if (min != null && domain.first() < (min!! as Instant)) (min!! as Instant) else domain.first()
                val newMax =
                    if (max != null && domain.last() > (max!! as Instant)) (max!! as Instant) else domain.last()

                var updatedZoom = zoom

                if (newMin > domain.first()) {
                    val diff = newMin.toEpochMilliseconds() - domain.first().toEpochMilliseconds()
                    val newMaxMoved = fromEpochMilliseconds(newMax.toEpochMilliseconds() + diff)
                    val updatedMax = if (max != null) fromEpochMilliseconds(
                        min(
                            (max!! as Instant).toEpochMilliseconds(),
                            newMaxMoved.toEpochMilliseconds()
                        )
                    ) else newMaxMoved

                    val finalScale = (_scale.copy() as ContinuousScale<Instant, Double>).also {
                        it.domain = listOf(newMin, updatedMax)
                    }
                    updatedZoom = finalScale.computeZoom((_scale as ContinuousScale<Instant, Double>))
                } else if (newMax < domain.last()) {
                    val diff = domain.last().toEpochMilliseconds() - newMax.toEpochMilliseconds()
                    val newMinMoved = fromEpochMilliseconds(newMin.toEpochMilliseconds() - diff)
                    val updatedMin = if (min != null) fromEpochMilliseconds(
                        max(
                            (min!! as Instant).toEpochMilliseconds(),
                            newMinMoved.toEpochMilliseconds()
                        )
                    ) else newMinMoved

                    val finalScale = (_scale.copy() as ContinuousScale<Instant, Double>).also {
                        it.domain = listOf(updatedMin, newMax)
                    }
                    updatedZoom = finalScale.computeZoom((_scale as ContinuousScale<Instant, Double>))
                }

                updatedZoom
            }
            else -> zoom
        }
    }

    /**
     * Apply a zoom to an Axis scale.
     */
    internal fun applyZoom(zoom: DirectedZoom) {
        if (dimension.isChildDimension) return

        // reset the scale
        _zoomScale = _scale

        // if zoom is "IDENTITY" just return
        if (zoom == DirectedZoom.IDENTITY) return

        // else we need to rescale (= recompute the scale) from the current zoom
        when (val sc = scale) {
            is ContinuousScale<VALUE, Double> -> {
                _zoomScale = zoom.rescale(sc as ContinuousScale<Double, Double>) as Scale<VALUE, Double>
                logger.verbose("updateScale from zoom[$zoom]:: ${sc.toDescription()}  ${_scale.toDescription()}")
            }
            is StrictlyContinuousRange<VALUE, Double> -> {
                val zoomedScale = _scale.copy() as StrictlyContinuousRange<VALUE, Double>
                zoomedScale.range =
                    StrictlyContinuous(zoom.apply(zoomedScale.range.start), zoom.apply(zoomedScale.range.end))
                _zoomScale = zoomedScale
                logger.verbose("updateScale from zoom[$zoom]:: ${sc.toDescription()}  ${_scale.toDescription()}")
            }
        }
    }

    private fun <DOMAIN, RANGE> Scale<DOMAIN, RANGE>.toDescription(): String {
        return when (this) {
            is ContinuousScale<*, *> -> "Continuous domain$domain range$range"
            is BandedScale<*> -> "BandedScale domain [${domain.firstOrNull()}...${domain.lastOrNull()}] range$range"
            else -> "$this"
        }
    }

    // Axis renderer -------------

    internal fun drawAxis(axisLayerBack: CElement, axisLayerFront: CElement) {
        if (isChild.not()) {

            // If we redraw an axis, it must have changed, so it may be necessary to recompute space and rotation
            _axisRenderer.computeLabelSpaceAndRotation()

            // Then draw the background of the axis (gridlines...)
            _axisRenderer.drawBackgroundElements(axisLayerBack)

            // Then draw the axis itself
            _axisRenderer.drawAxis(axisLayerFront)
        }
    }


    // Ticks -----

    /**
     * Specify a [List] of VALUES for the ticks, if none is defined, this [Axis] will compute its own ticks.
     */
    @Deprecated("Use the [dimension.formatter] to provide custom ticks.")
    public var ticks: List<VALUE>? = null

    /**
     * Default implementation to retrieve ticks: if the scale is [Tickable] use
     * the scale implementation.
     *
     */
    private fun defaultTicksImpl(): List<VALUE> {
        return when (val scale = scale) {
            is BandedScale -> getDefaultDiscreTicks(scale)
            is Tickable<*> -> {
                try {
                    scale as ContinuousRangeScale
                    val start = scale.range.first()
                    val end = scale.range.last()
                    getDefaultContinuousTicks(scale, abs(end - start))
                } catch (e: Exception) {
                    getDefaultContinuousTicks(scale, 600.0)
                }
            }
            else -> listOf()
        }
    }

    private fun getDefaultContinuousTicks(scale: Tickable<*>, rangeSize: Double): List<VALUE> =
        (scale as Tickable<VALUE>).ticks(tickCount ?: smartTickCount(rangeSize, orientation))

    private fun getDefaultDiscreTicks(scale: BandedScale<VALUE>): List<VALUE> {
        return scale.ticks()
    }

    /**
     * Specify the lambda to retrieve ticks.
     */
    @Deprecated("Use the [dimension.formatter] to provide custom ticks.")
    public fun ticks(getTicks: () -> List<VALUE>) {
//        _getTicks = getTicks
    }


    internal fun smartTickCount(rangeSize: Double, orientation: Orientation): Int =
        tickCount ?:
        if (orientation.isHorizontal) {
            with(abs(rangeSize)) {
                when {
//                this <= 220  -> 2
                    this <= 600 -> 5
                    this <= 1200 -> 10
                    this <= 2000 -> 15
                    else -> 20
                }
            }
        } else {
            with(abs(rangeSize)) {
                when {
//                this <= 220  -> 2
                    this <= 400 -> 5
                    this <= 800 -> 10
                    this <= 1200 -> 15
                    else -> 20
                }
            }
        }


}

