package io.data2viz.charts.chart.mark

import io.data2viz.charts.chart.Chart
import io.data2viz.charts.chart.mark.internal.BoxPlotMarkImpl
import io.data2viz.charts.core.Dataset
import io.data2viz.charts.core.Datum
import io.data2viz.charts.core.Orientation
import io.data2viz.charts.dimension.Dimension
import io.data2viz.charts.dimension.Discrete
import io.data2viz.charts.dimension.Quantitative
import io.data2viz.geom.Point

/**
 * A boxplot mark "context".
 * Each Point if it exists, is placed at the "center" of the boxplot, and is relative to the current [DrawingZone].
 * The bandwidth is the expected "thickness" of the boxplot candle.
 */
public data class BoxPlotContext(
    val orientation: Orientation,
    val mean: Point?,
    val median: Point?,
    val minimum: Point,
    val maximum: Point,
    val lowerQuartile: Point,
    val upperQuartile: Point,
    val outliers: List<Point>,
    val bandwidth: Double
)

/**
 * Add a vertical [BoxPlotMark] to the current [Chart].
 *
 * See [Box plot](https://en.wikipedia.org/wiki/Box_plot) for more info.
 *
 * @param x The [Discrete] dimension for the X axis.
 * @param y The [Quantitative] dimension for the Y axis.
 * @param init The initialization lambda for this [Mark].
 */
public fun <DOMAIN, XVAL> Chart<DOMAIN>.boxPlot(
    x: Discrete<DOMAIN, XVAL>,
    y: Quantitative<DOMAIN>,
    init: BoxPlotMark<DOMAIN, XVAL, Double?>.() -> Unit = {}
): Unit = addMark { markIndex -> BoxPlotMarkImpl(markIndex, dataset, x, y, true, init, config.events, config.mark, config.x, config.y) }

/**
 * Add a horizontal [BoxPlotMark] to the current [Chart].
 *
 * See [Box plot](https://en.wikipedia.org/wiki/Box_plot) for more info.
 *
 * @param x The [Quantitative] dimension for the X axis.
 * @param y The [Discrete] dimension for the Y axis.
 * @param init The initialization lambda for this [Mark].
 */
public fun <DOMAIN, YVAL> Chart<DOMAIN>.boxPlot(
    x: Quantitative<DOMAIN>,
    y: Discrete<DOMAIN, YVAL>,
    init: BoxPlotMark<DOMAIN, Double?, YVAL>.() -> Unit = {}
): Unit = addMark { markIndex -> BoxPlotMarkImpl(markIndex, dataset, x, y, false, init, config.events, config.mark, config.x, config.y) }

/**
 * Draw BoxPlot (or candle) for a given [Dataset] and X and Y [Dimension]s.
 */
public interface BoxPlotMark<DOMAIN, XVAL, YVAL> :
    MarkXY<DOMAIN, XVAL, YVAL>,
    HasMarker<DOMAIN>,
    HasFill<DOMAIN>,
    HasSize<DOMAIN>,
    HasStrokeColor<DOMAIN>,
    HasBoxPlotDecorator<DOMAIN>,
    HasBoxPlotHighlightDecorator<DOMAIN>,
    HasBoxPlotSelectionDecorator<DOMAIN>,
    HasStrokeWidth<DOMAIN>,
    HasCustomTooltip<DOMAIN>,
    HasHighlightMode,
    HasSelectionMode,
    HasLegendDecorator<DOMAIN>
{

    /**
     * The "minimum" value accessor, use this to provide the "minimum" value for each of your domain object.
     *
     * See [Elements of a box plot](https://en.wikipedia.org/wiki/Box_plot#Elements_of_a_box_plot) for more info.
     */
    public var minimum: Datum<DOMAIN>.() -> Double

    /**
     * The "maximum" value accessor, use this to provide the "maximum" value for each of your domain object.
     *
     * See [Elements of a box plot](https://en.wikipedia.org/wiki/Box_plot#Elements_of_a_box_plot) for more info.
     */
    public var maximum: Datum<DOMAIN>.() -> Double

    /**
     * The "lower Quartile" or "First quartile" value accessor, use this to provide the value for each of your
     * domain object.
     *
     * See [Elements of a box plot](https://en.wikipedia.org/wiki/Box_plot#Elements_of_a_box_plot) for more info.
     */
    public var lowerQuartile: Datum<DOMAIN>.() -> Double

    /**
     * The "upper Quartile" or "Third  quartile" value accessor, use this to provide the value for each of your
     * domain object.
     *
     * See [Elements of a box plot](https://en.wikipedia.org/wiki/Box_plot#Elements_of_a_box_plot) for more info.
     */
    public var upperQuartile: Datum<DOMAIN>.() -> Double

    /**
     * The "mean" value accessor, use this to provide the value for each of your domain object.
     * This is null by default if you do not want to display the mean.
     */
    public var mean: (Datum<DOMAIN>.() -> Double)?

    /**
     * The "outliers" value(s) accessor, use this to provide the "outliers" value(s) as a [List] for each of your
     * domain object.
     *
     * See [Elements of a box plot](https://en.wikipedia.org/wiki/Box_plot#Elements_of_a_box_plot) for more info.
     */
    public var outliers: Datum<DOMAIN>.() -> List<Double>

    /**
     * The box plot shape for each Datum.
     * Default to { BoxPlotShape.Bar }.
     */
    public var shape: Datum<DOMAIN>.() -> BoxPlotShape

}

@Deprecated(
    "Replace with BoxPlotMark.",
    ReplaceWith("BoxPlotMark<DOMAIN, XVAL, YVAL>", "io.data2viz.charts.chart.mark.BoxPlotMark")
)
public typealias BoxPlot<DOMAIN, XVAL, YVAL> = BoxPlotMark<DOMAIN, XVAL, YVAL>
