package io.data2viz.charts.chart.mark

import io.data2viz.charts.chart.Chart
import io.data2viz.charts.chart.mark.domainSpecific.PriceMovement
import io.data2viz.charts.chart.mark.internal.CandleStickMarkImpl
import io.data2viz.charts.core.Dataset
import io.data2viz.charts.core.Datum
import io.data2viz.charts.core.Orientation
import io.data2viz.charts.dimension.Dimension
import io.data2viz.charts.dimension.Quantitative
import io.data2viz.charts.dimension.Temporal
import io.data2viz.geom.Point
import kotlinx.datetime.Instant

/**
 * A candlestick mark "context".
 * Each Point, is placed at the "center" of the candlestick, and is relative to the current [DrawingZone].
 * The bandwidth is the expected "thickness" of the candle.
 */
public data class CandleStickContext(
    val orientation: Orientation,
    val open: Point,
    val close: Point,
    val high: Point,
    val low: Point,
    val bandwidth: Double
)

/**
 * Add a vertical [CandleStickMark] to the current [Chart].
 *
 * See [Candlestick](https://en.wikipedia.org/wiki/Candlestick_chart) for more info.
 *
 * @param x The [Temporal] dimension for the X axis.
 * @param y The [Quantitative] dimension for the Y axis.
 * @param priceMovement The [PriceMovement] value accessor.
 * @param init The initialization lambda for this [Mark].
 */
public fun <DOMAIN> Chart<DOMAIN>.candleStick(
    x: Temporal<DOMAIN>,
    y: Quantitative<DOMAIN>,
    priceMovement: Datum<DOMAIN>.() -> PriceMovement?,
    init: CandleStickMark<DOMAIN, Instant, Double?>.() -> Unit = {}
): Unit = addMark { markIndex -> CandleStickMarkImpl(
    markIndex, dataset, x, y, priceMovement, true, init, config.events, config.mark, config.x, config.y
) }

/**
 * Add a horizontal [CandleStickMark] to the current [Chart].
 *
 * See [Candlestick](https://en.wikipedia.org/wiki/Candlestick_chart) for more info.
 *
 * @param x The [Temporal] dimension for the X axis.
 * @param y The [Quantitative] dimension for the Y axis.
 * @param priceMovement The [PriceMovement] value accessor.
 * @param init The initialization lambda for this [Mark].
 */
public fun <DOMAIN> Chart<DOMAIN>.candleStick(
    x: Quantitative<DOMAIN>,
    y: Temporal<DOMAIN>,
    priceMovement: Datum<DOMAIN>.() -> PriceMovement?,
    init: CandleStickMark<DOMAIN, Double?, Instant>.() -> Unit = {}
): Unit = addMark { markIndex -> CandleStickMarkImpl(
    markIndex, dataset,  x, y, priceMovement, false, init, config.events, config.mark, config.x, config.y
) }

/**
 * Draw CandleStick for a given [Dataset] and X and Y [Dimension]s.
 */
public interface CandleStickMark<DOMAIN, XVAL, YVAL> :
    MarkXY<DOMAIN, XVAL, YVAL>,
    HasFill<DOMAIN>,
    HasStrokeColor<DOMAIN>,
    HasCandleStickDecorator<DOMAIN>,
    HasCandleStickHighlightDecorator<DOMAIN>,
    HasCandleStickSelectionDecorator<DOMAIN>,
    HasStrokeWidth<DOMAIN>,
    HasCustomTooltip<DOMAIN>,
    HasHighlightMode,
    HasSelectionMode
{
    /**
     * The gap is the fixed positive value in "pixels" between 2 candles.
     * Set gap to 0 (default value) for candles to stick together, set it to a higher value to increase readability.
     * Note that this works in addition to the bandwidth property of the [CandleStickContext] and so, should
     * be used with care as it implies to display incorrect values if set >0.
     *
     * @see CandleStickContext
     */
    public var gap: Int
}
