package io.data2viz.charts.chart.mark

import io.data2viz.charts.chart.Chart
import io.data2viz.charts.chart.mark.internal.*
import io.data2viz.charts.config.AxisConfig
import io.data2viz.charts.config.EventConfig
import io.data2viz.charts.config.MarkConfig
import io.data2viz.charts.core.Dataset
import io.data2viz.charts.dimension.Dimension

/**
 * Instantiate a new [LineMark].
 *
 * See [Line chart](https://en.wikipedia.org/wiki/Line_chart) for more info.
 *
 * @param dataset Your chart's [Dataset]
 * @param x The [Dimension] for the X axis, this dimension can be continuous or discrete
 * @param y The [Dimension] for the Y axis, this dimension can be continuous or discrete
 * @param init The initialization lambda for this [Mark]
 * @param config The [MarkConfig] to configure this [Mark]
 * @param xAxisConfig The [AxisConfig] to configure this mark's X axis
 * @param yAxisConfig The [AxisConfig] to configure this mark's Y axis
 */
public fun <DOMAIN, XVAL, YVAL> lineMark(
    markIndex: Int,
    dataset: Dataset<DOMAIN>,
    x: Dimension<DOMAIN, XVAL>,
    y: Dimension<DOMAIN, YVAL>,
    init: LineMark<DOMAIN, XVAL, YVAL>.() -> Unit,
    eventConfig: EventConfig,
    config: MarkConfig,
    xAxisConfig: AxisConfig,
    yAxisConfig: AxisConfig
): LineMark<DOMAIN, XVAL, YVAL> = LineMarkImpl(markIndex, dataset, x, y, init, eventConfig, config, xAxisConfig, yAxisConfig)

/**
 * Add a [LineMark] to the current [Chart].
 *
 * See [Line chart](https://en.wikipedia.org/wiki/Line_chart) for more info.
 *
 * @param x The [Dimension] for the X axis, this dimension can be continuous or discrete.
 * @param y The [Dimension] for the Y axis, this dimension can be continuous or discrete.
 * @param init The initialization lambda for this [Mark].
 */
public fun <DOMAIN, XVAL, YVAL> Chart<DOMAIN>.line(
    x: Dimension<DOMAIN, XVAL>,
    y: Dimension<DOMAIN, YVAL>,
    init: LineMark<DOMAIN, XVAL, YVAL>.() -> Unit = {}
): Unit = addMark { markIndex -> lineMark(markIndex, dataset, x, y, init, config.events, config.mark, config.x, config.y) }


/**
 * Draw lines connecting each elements for a given [Dataset] and X and Y [Dimension]s.
 */
public interface LineMark<DOMAIN, XVAL, YVAL> :

    HasMarkDecorator<DOMAIN>,
    HasHighlightDecorator<DOMAIN>,
    HasSelectionDecorator<DOMAIN>,
    HasEventZones<DOMAIN>,
    HasLegendDecorator<DOMAIN>,
    HasCurve,
    HasSize<DOMAIN>,
    HasMarker<DOMAIN>,
    HasStrokeColor<DOMAIN>,
    HasStrokeWidth<DOMAIN>,
    HasStrokeLine<DOMAIN>,
    HasHighlightMode,
    HasSelectionMode,
    MarkXY<DOMAIN, XVAL, YVAL> {

    // TODO HasJoinMissingValues ?
    /**
     * Does this [LineMark] link 2 points over a missing value or does it cut the line?
     *
     * When a VALUE is missing in the X or Y [Dimension] (meaning that the DOMAIN object has a null VALUE in this
     * [Dimension]) then the point in the [Path] will be skipped.
     *
     * - use [joinMissingValues] = true to link the neighbour points.
     * - use [joinMissingValues] = false (default) to cut the line on each missing point.
     */
    public var joinMissingValues: Boolean

}

@Deprecated(
    "Replace with LineMark.",
    ReplaceWith("LineMark<DOMAIN, XVAL, YVAL>", "io.data2viz.charts.chart.mark.LineMark")
)
public typealias Line<DOMAIN, XVAL, YVAL> = LineMark<DOMAIN, XVAL, YVAL>
