package io.data2viz.charts.chart.mark

import io.data2viz.charts.core.Dataset
import io.data2viz.charts.core.Datum
import io.data2viz.charts.event.EventZone
import io.data2viz.charts.layout.DrawingZone

/**
 * A [Mark] is a way to represent a data as visual elements (signs).
 *
 * For example a [LineMark] will display a curve between all elements, a [PlotMarkImpl] will display a sign for
 * each element etc, a [ColumnMark] will display vertical bars for each element...
 */
public interface Mark<DOMAIN> {

    /**
     * The [prepare] function is called once on the initialization of mark and on data changed.
     *
     * You can use this to indicate some requirements by using [require] or to compute some elements that do
     * not change until data change.
     *
     * @param [dataset] the chart's [Dataset]
     */
    public fun prepare(dataset: Dataset<DOMAIN>) { }

    /**
     * Draw a sign on the [DrawingZone] for each elements of a given dataset.
     *
     * Use the positions ([MarkX.x], [MarkY.y]...) lambdas to draw signs on the expected positions.
     *
     * Notes :
     *  - You usually want to clear the [DrawingZone] before drawing new signs.
     *  - Depending on the positions lambdas, some signs may be drawn out of the drawing zone and they may be skipped.
     *
     * @param [drawingZone] the zone allocated by the [Chart] component for the [Mark] to draw itself
     * @param [dataset] the chart's [Dataset]
     */
    public fun drawMark(drawingZone: DrawingZone, dataset: Dataset<DOMAIN>) { }


    /**
     * Draw a highlighted sign on the [DrawingZone] for each elements of a given dataset.
     *
     * Use the positions ([MarkX.x], [MarkY.y]...) lambdas to draw signs on the expected positions.
     *
     * Notes :
     *  - You usually want to clear the [DrawingZone] before drawing new signs.
     *  - Depending on the positions lambdas, some signs may be drawn out of the drawing zone and they may be skipped.
     *
     * @param [drawingZone] the zone allocated by the [Chart] component for the [Mark] to draw itself
     * @param [highlightedData] a collection of highlighted [Datum] objects
     */
    public fun drawHighlight(drawingZone: DrawingZone, highlightedData: Collection<Datum<DOMAIN>>) { }


    /**
     * Draw a selected sign on the [DrawingZone] for each elements of a given dataset.
     *
     * Use the positions ([MarkX.x], [MarkY.y]...) lambdas to draw signs on the expected positions.
     *
     * Notes :
     *  - You usually want to clear the [DrawingZone] before drawing new signs.
     *  - Depending on the positions lambdas, some signs may be drawn out of the drawing zone and they may be skipped.
     *
     * @param [drawingZone] the zone allocated by the [Chart] component for the [Mark] to draw itself
     * @param [selectedData] a collection of selected [Datum] objects
     */
    public fun drawSelection(drawingZone: DrawingZone, selectedData: Collection<Datum<DOMAIN>>) { }


    /**
     * Get all of interactive zones for this [Mark] on a given [DrawingZone].
     *
     * When drawing a [Mark] some zones need to react to user interactions (hovering, clicking...) and some may not,
     * to know this the [Chart] will call this [getEventZones] function to retrieve a collection of [EventZone] of
     * various shapes.
     * The [DrawingZone] can be used to "limit" the [EventZone] to a given screen space.
     *
     * Note: when creating custom marks, it is highly recommended to create and store these [EventZone]s,
     * generally this can be done in the [drawMark] function, to avoid recreating them each time.
     *
     * @param [drawingZone] the zone allocated by the [Chart] component for the [Mark] to draw itself
     * @param [dataset] the chart's [Dataset]
     * @return a Collection of [EventZone] drawn by this mark.
     */
    public fun getEventZones(drawingZone: DrawingZone, dataset: Dataset<DOMAIN>): Collection<EventZone<DOMAIN>> = listOf()


}
