package io.data2viz.charts.chart.mark.domainSpecific

import kotlinx.datetime.Instant
import kotlin.time.Duration
import io.data2viz.charts.chart.mark.CandleStickMark

/**
 * Represents a price movement in a financial market.
 *
 * This data class is specific to the [CandleStickMark] and contains information to display.
 * Use this data class to build a collection of PriceMovement from your own list of domain objects, for example
 *      fun MyDomainObjectClass.toPriceMovement(): PriceMovement
 *
 * @property timestamp The timestamp of the price movement.
 * @property interval The duration of the interval.
 * @property open The open price as a double value.
 * @property close The closing price value.
 * @property high The highest price value.
 * @property low The lowest price observed.
 * @property volume The volume of the price movement.
 * @property turnover The overall value of all transactions for the given interval.
 */
public data class PriceMovement (

    /**
     * Represents the timestamp of a price movement in a financial market.
     */
    public val timestamp: Instant,

    /**
     * Represents the interval between two price movements in a financial market.
     */
    public val interval: Duration,

    /**
     * Represents the open price of a price movement in a financial market.
     *
     * The open price is the first price at which a financial instrument is traded during a given interval.
     * It provides information about the initial level of demand and supply for the instrument.
     */
    public val open: Double,

    /**
     * Represents the closing price of a price movement in a financial market.
     */
    public val close: Double,

    /**
     * Represents the highest price during a given interval in a financial market.
     */
    public val high: Double,

    /**
     * Represents the lowest price observed in a price movement of a financial market.
     */
    public val low: Double,

    /**
     * Represents the volume of a price movement in a financial market.
     * Volume refers to the total amount of shares, contracts, or lots traded in a given period.
     */
    public val volume: Double,

    /**
     * Represents the total value of all transactions within a specific time interval.
     */
    public val turnover: Double
)
