package io.data2viz.charts.config

import io.data2viz.charts.chart.internal.CursorImpl
import io.data2viz.viz.HasStroke
import io.data2viz.charts.core.SelectedDatum
import io.data2viz.charts.core.CursorDisplay
import io.data2viz.charts.core.CursorMode
import io.data2viz.charts.core.Font
import io.data2viz.charts.core.CursorType
import io.data2viz.charts.core.Stroke
import io.data2viz.charts.core.ExperimentalCharts
import io.data2viz.charts.layout.DrawingZone
import io.data2viz.geom.Point

/**
 * A decorator lambda is called for the cursor, use this to draw custom cursors.
 *
 * @param selectedData the current [SelectedDatum]
 * @param position the expected position where the cursor should appear (relative to the [drawingZone] coordinates)
 * @param drawingZone the drawing zone on which to draw the decorator
 */
public typealias CursorDecorator<DOMAIN> = (selectedData: SelectedDatum<DOMAIN>?, position: Point, drawingZone: DrawingZone) -> Unit

/**
 * The configuration interface for a [CursorImpl].
 */
public interface CursorConfig : Stroke, Font {

    /**
     * Show / hide the Cursor (defaults to false).
     */
    public var show: Boolean

    /**
     * Type of Cursor: horizontal, vertical or crosshair (defaults to "Auto").
     */
    public var type: CursorType

    /**
     * Cursor can display additional information on axes (defaults to "None").
     */
    public var display: CursorDisplay

    /**
     * Mode of Cursor positioning: stick on X/Y or free (defaults to "StickOnXY").
     */
    public var mode: CursorMode

    /**
     * The stroke line (solid or dashed line), defaults to doubleArrayOf(3.0, 3.0).
     *
     * [1.0] =      --------------
     * [1.0, 1.0] = - - - - - - --
     * [2.0, 1.0] = -- -- -- -- --
     * [1.0, 3.0] = -   -   -   -
     *
     * @see HasStroke
     */
    public var strokeLine: DoubleArray


    /**
     * The cursor decorator lambda will be called on cursor display, use this to draw custom cursors.
     */
    @ExperimentalCharts
    public var cursorDecorator: CursorDecorator<*>?

}
