package io.data2viz.charts.core

import io.data2viz.charts.deprecate.MoveToInternal
import io.data2viz.charts.dimension.Constant
import io.data2viz.charts.dimension.DiscreteDimension


internal data class Series(val index:Int) {
    var elements = 0
}


/**
 * Container for the Chart data, gives access to various properties.
 * @see [Datum]
 */
@MoveToInternal
public class Dataset<DOMAIN> internal constructor(internal val dat: List<DOMAIN>) {

    private val _data = computeInternalData(dat)
    internal var domainToDatum: Map<DOMAIN, Datum<DOMAIN>> = computeDomainToDatum()

    internal var seriesSet = false

    internal var series: DiscreteDimension<DOMAIN, *> = Constant(0)
        set(value) {
            field = value
            seriesSet = true

            val allSeries:MutableMap<Any?, Series> = mutableMapOf()
            var indexOfSeries = 0
            data.forEach { datum ->
                val seriesKey = series(datum)
                val seriesObject = allSeries.getOrPut(seriesKey) { Series(indexOfSeries++) }
                datum._indexOfSeries = seriesObject.index
                datum._indexInSeries = seriesObject.elements++
            }

            _dataBySeriesComputed = false               // needed to force refresh of dataBySeries
            singleSeries = dataBySeries.size == 1
        }

    private lateinit var _dataBySeries: List<List<Datum<DOMAIN>>>
    private lateinit var _dataBySeriesMap: Map<*, List<Datum<DOMAIN>>>

    private var _dataBySeriesComputed = false
    private var _dataBySeriesMapComputed = false


    /**
     * A List<List<Datum>>: the first index is the element's series index, second index is the element's
     * index in the series.
     *
     */
    public val dataBySeries: List<List<Datum<DOMAIN>>>
        get() {
            if (!_dataBySeriesComputed) {
                _dataBySeriesComputed = true
                _dataBySeries = data.groupBy { it.indexOfSeries }.map { it.value }
            }
            return _dataBySeries
        }


    /**
     * A List<List<Datum>>: the first index is the element's series index, second index is the element's
     * index in the series.
     */
    internal val dataBySeriesMap: Map<*, List<Datum<DOMAIN>>>
        get() {
            if (!_dataBySeriesMapComputed) {
                _dataBySeriesMapComputed = true
                _dataBySeriesMap = data.groupBy { series(it) }
            }
            return _dataBySeriesMap
        }

    internal var singleSeries: Boolean = true

    /**
     * A [List] of [Datum] objects created from the original collection of DOMAIN objects.
     */
    public val data: List<Datum<DOMAIN>>
        get() = _data

    private fun computeDomainToDatum() = _data.map { datum -> Pair(datum.domain, datum) }.toMap()
    private fun computeInternalData(dat: List<DOMAIN>) = dat.mapIndexed { index, domain -> Datum(this, domain, index) }

}
