package io.data2viz.charts.core

import io.data2viz.format.Locale
import io.data2viz.format.Type
import io.data2viz.format.formatter
import io.data2viz.format.precisionRound
import kotlin.math.*

private fun Double?.nullFormat(formatter: (Double) -> String ) = this?.let { formatter(it) } ?: ""

/**
 * Format the Double? to a rounded integer.
 * PI.formatToInteger() = "3"
 */
public fun Double?.formatToInteger(locale: Locale = Locale()): String = nullFormat(locale.formatter(type = Type.DECIMAL_ROUNDED))

/**
 * Format the Double? to SI notation.
 * PI.formatToSI() = "3.14159"
 */
public fun Double?.formatToSI(locale: Locale = Locale()): String =             nullFormat(locale.formatter(type = Type.DECIMAL_WITH_SI, precision = 3))

/**
 * Format the Double? to either decimal or exponent notation, rounded to significant digits.
 * PI.formatToDecimal() = "3.14159"
 */
public fun Double?.formatToDecimal(locale: Locale = Locale()): String = nullFormat(locale.formatter(type = Type.DECIMAL_OR_EXPONENT, precision = 6))

/**
 * Format the Double? to exponent notation.
 * PI.formatToSI() = "3.141593e+0"
 */
public fun Double?.formatToExponent(locale: Locale = Locale()): String =       nullFormat(locale.formatter(type = Type.EXPONENT))

/**
 * Format the Double? to fixed point notation.
 * PI.formatToFixedPoint() = "3.141593"
 */
public fun Double?.formatToFixedPoint(locale: Locale = Locale()): String =     nullFormat(locale.formatter(type = Type.FIXED_POINT))

/**
 * Format the Double? to percentage notation.
 * PI.formatToPercent() = "314.159265%"
 */
public fun Double?.formatToPercent(locale: Locale = Locale()): String =        nullFormat(locale.formatter(type = Type.PERCENT, precision = 2))

/**
 * Format the Double? to percentage notation rounded to significant digits.
 * PI.formatToPercentInteger() = "314.159%"
 */
public fun Double?.formatToPercentInteger(locale: Locale = Locale()): String = nullFormat(locale.formatter(type = Type.PERCENT_ROUNDED, precision = 0))

/**
 * Format the Double? to binary notation.
 * PI.formatToBinary() = "11"
 */
public fun Double?.formatToBinary(locale: Locale = Locale()): String =         nullFormat(locale.formatter(type = Type.BINARY))

/**
 * Format the Double? to octal notation.
 * PI.formatToOctal() = "3"
 */
public fun Double?.formatToOctal(locale: Locale = Locale()): String =          nullFormat(locale.formatter(type = Type.OCTAL))

/**
 * Format the Double? to hexadecimal notation.
 * PI.formatToHexadecimal() = "3"
 */
public fun Double?.formatToHexadecimal(locale: Locale = Locale()): String = nullFormat(locale.formatter(type = Type.HEX_UPPERCASE))

/**
 * "Smart format" the Double.
 * PI.formatSmart() = "3.14159"
 */
public fun Double?.formatSmartDouble(locale: Locale): String {
    val smartFormatter: (Double) -> String = {
//        val precision = if (it.toInt().toDouble() == it) 0 else if ((it*100).toInt().toDouble() == it*100) 2 else 6
        if (it != .0 && (abs(it) >= 10000 || abs(it) <= 0.0001)) {
            it.formatToSI(locale)
        } else {
            val isInteger = abs(it % 1) < 0.000001
            if (isInteger) it.formatToInteger(locale) else it.formatToDecimal(locale)
        }
    }
    return nullFormat(smartFormatter)
}

// TODO build the smartFormatter "once and for all" (but each time the axis domain change we must recreate a new one)
/**
 * "Smart format"
 */
public fun Double?.formatSmart(locale: Locale, doubleRange: Double): String {
    val invDouble = abs(round(1.0 / doubleRange))
    val precision = if (invDouble <= 1.0) 0 else ceil(log10(invDouble)).toInt()
    val smartFormatter: (Double) -> String = {
        if (precision == 0 && doubleRange > 10000)
            locale.formatter(type = Type.DECIMAL_OR_EXPONENT, precision = precision)(it)
        else locale.formatter(type = Type.FIXED_POINT, precision = precision)(it)
    }
    return nullFormat(smartFormatter)
}

/**
 * "smart format" a Double knowing the max value to be displayed and the minimal step between 2 values.
 * This is used for axes, to create a smart value formatter.
 */
public fun Double?.formatSmartDouble(locale: Locale, step: Double, max: Double): String =
    nullFormat { locale.formatter(".${precisionRound(step, max)}s")(it) }

/**
 * List of formatters for Double.
 */
public object DoubleFormatters {

    /**
     * Format the Double to a rounded integer.
     * DoubleFormatters.integer(PI) = "3"
     */
//    public fun integer(l:Locale): (Double?) -> String = { p1: Double? ->
//        p1.formatToInteger(l)
//    }

    /**
     * Format the Double to decimal notation, rounded to significant digits.
     * DoubleFormatters.decimal(PI) = "3.14159"
     */
//    public object decimal: (Double?, Locale) -> String {
//        override fun invoke(p1: Double?, l: Locale): String = p1.formatToDecimal(l)
//    }

    /**
     * Format the Double to exponent notation.
     * DoubleFormatters.SI(PI) = "3.141593e+0"
     */
//    public object SI: (Double?, Locale) -> String {
//        override fun invoke(p1: Double?, l: Locale): String = p1.formatToSI(l)
//    }

    /**
     * Format the Double to exponent notation.
     * DoubleFormatters.exponent(PI) = "3.141593e+0"
     */
//    public object exponent: (Double?, Locale) -> String {
//        override fun invoke(p1: Double?, l: Locale): String = p1.formatToExponent(l)
//    }

    /**
     * Format the Double to fixed point notation.
     * DoubleFormatters.fixedPoint(PI) = "3.141593"
     */
//    public object fixedPoint: (Double?, Locale) -> String {
//        override fun invoke(p1: Double?, l: Locale): String = p1.formatToFixedPoint(l)
//    }

    /**
     * Format the Double to percentage notation.
     * DoubleFormatters.percent(PI) = "314.159265%"
     */
//    public object percent: (Double?, Locale) -> String {
//        override fun invoke(p1: Double?, l: Locale): String = p1.formatToPercent(l)
//    }

    /**
     * Format the Double to percentage notation rounded to significant digits.
     * DoubleFormatters.percentInteger(PI) = "314.159%"
     */
//    public fun percentInteger(l: Locale): (Double?) -> String = { p1: Double? ->
//        p1.formatToPercentInteger(l)
//    }

    /**
     * Format the Double to binary notation.
     * DoubleFormatters.binary(PI) = "11"
     */
//    public object binary: (Double?, Locale) -> String {
//        override fun invoke(p1: Double?, l: Locale): String = p1.formatToBinary(l)
//    }

    /**
     * Format the Double to octal notation.
     * DoubleFormatters.octal(PI) = "3"
     */
//    public object octal: (Double?, Locale) -> String {
//        override fun invoke(p1: Double?, l: Locale): String = p1.formatToOctal(l)
//    }

    /**
     * Format the Double to hexadecimal notation.
     * DoubleFormatters.hexadecimal(PI) = "3"
     */
//    public object hexadecimal: (Double?, Locale) -> String {
//        override fun invoke(p1: Double?, l: Locale): String = p1.formatToHexadecimal(l)
//    }

    /**
     * "Smart format" the Double.
     * DoubleFormatters.smart(PI) = "3.14159"
     */
//    public object smart: (Double?, Locale) -> String {
//        override fun invoke(p1: Double?, l: Locale): String = p1.formatSmart(l)
//    }
}
