package io.data2viz.charts.dimension

import io.data2viz.charts.config.AxisConfig
import io.data2viz.charts.core.Datum

/**
 * A dimension is an object that explains how a "DOMAIN" object is converted to a "VALUE" (through the Dimension's
 * [accessor]) and how this "VALUE" should be interpreted: numeric, temporal or discrete?
 *
 * For example, your DOMAIN is a data class:
 *      data class WeatherRecord(val date:Date, val weather:WeatherType, val temperature:Double)
 *
 * You can encode these different properties through [Dimension]s:
 *
 *  - val temperatureDimension = quantitative( { domain.temperature } )
 *  - val weatherDimension = discrete( { domain.weatherType } )
 *  - val timeDimension = temporal( { domain.date } )
 *
 * The Dimension constructor takes mandatory information:
 *  - the type of your value (discrete, temporal, quantitative...)
 *  - how to access your value (the [accessor])
 *
 * Dimension also has a DSL to specify various additional elements like:
 *  - the name of your value
 *  - its formatting function
 *
 * For example:
 * ```kotlin
 * val temperatureDimension = quantitative( { domain.temperature } ) {
 *      name = "Mean daily temperature (in Celsius)"            // the name (used on axes, tooltip or in legend)
 *      format = { "$this °C" }                                 // the formatting function
 * }
 * ```
 */
public sealed class Dimension<DOMAIN, VALUE>(

    /**
     * The "accessor" is the lambda used to transform a DOMAIN object into a VALUE.
     */
    public val accessor: Datum<DOMAIN>.() -> VALUE,
    internal val parentDimension: Dimension<DOMAIN,VALUE>? = null,
) {

    /*************************************** DSL ***************************************/

    /**
     * The name of this [Dimension], this name is used as the title of the [Axis] based on it.
     */
    public var name: String? = null

    /**
     * The formatter for the VALUE this [Dimension] can generate.
     */
    public open var formatter: (VALUE.() -> String)? = null


    /************************* DEPRECATED ********************************************/

    // TODO remove when axis() is removed
    internal var axisConfigInitializer: (AxisConfig.() -> Unit)? = null

    @Deprecated("Set the Axis configuration properties in the mark.x or mark.y instead", level = DeprecationLevel.ERROR)
    public fun axis(init: AxisConfig.() -> Unit) {
        axisConfigInitializer = init
    }

    @Deprecated("Remove Dimension.min, use mark.x.min or mark.y.min instead", level = DeprecationLevel.ERROR)
    public var min: VALUE? = null

    @Deprecated("Remove Dimension.max, use mark.x.max or mark.y.max instead", level = DeprecationLevel.ERROR)
    public var max: VALUE? = null


    /************************************ INTERNALS ************************************/

    /**
     * Indicates that the current dimension is linked to a parent one
     */
    internal var isChildDimension: Boolean = parentDimension != null

    internal var forcedTickCount: Int? = null


    /**
     * Is this [Dimension] considered "discrete"?
     */
    public abstract val isDiscrete: Boolean

    /**
     * Is this [Dimension] considered "continuous"?
     */
    public val isContinuous: Boolean
        get() = !isDiscrete

    /**
     * Calling Dimension([Datum]) is the same as calling its [accessor].
     */
    public operator fun invoke(datum: Datum<DOMAIN>): VALUE = accessor(datum)

    override fun toString(): String = name ?: super.toString()

}
