package io.data2viz.charts.dimension

import io.data2viz.charts.core.Datum


/**
 * [Quantitative] is a [Dimension] that is used to manage continuous numeric values (as [Double])
 * It is build by giving an "accessor" lambda that returns a [Double] object from a given [Datum] of a <DOMAIN>.
 *
 * Use the [Chart.quantitative] function to build a [Quantitative] dimension.
 *
 * Examples:
 *  - val sizeDimension = quantitative( { domain.size } )
 *  - val speedDimension = quantitative( { domain.speed } )
 */
public class Quantitative<DOMAIN>(
    accessor: Datum<DOMAIN>.() -> Double?,
    parentDimension: Quantitative<DOMAIN>? = null,
    init: Quantitative<DOMAIN>.() -> Unit = {}
) : ContinuousDimension<DOMAIN, Double?>(
    accessor = accessor,
    parentDimension = parentDimension
) {

    /************************************ INTERNALS ************************************/

    init {
        init()
    }


    /**
     * Creates a child [Dimension] with the given accessor, use child dimensions to share axes with
     * different properties.
     *
     * Note: if current dimension is already a "child" dimension, attach this dimension to the current's parent.
     *
     * @param accessor This child [Dimension] accessor
     * @param init The initialization lambda for this [Dimension]
     */
    public fun child(
        accessor: Datum<DOMAIN>.() -> Double?,
        init: Quantitative<DOMAIN>.() -> Unit = {}
    ): Quantitative<DOMAIN> {
        return Quantitative(accessor, (parentDimension ?: this) as Quantitative<DOMAIN>, init)
    }
}
