package io.data2viz.charts.event

import io.data2viz.charts.chart.Chart
import io.data2viz.charts.chart.mark.RelativePoint
import io.data2viz.charts.core.Datum
import io.data2viz.charts.core.SelectedDatum
import io.data2viz.geom.Rect

/**
 * EventContext contains the event type and the event context (elements that are impacted, ratios, size, ...)
 */
public interface EventContext<DOMAIN> : UserEvent {

    /**
     * A reference to the chart.
     */
    public val chart: Chart<DOMAIN>

    /**
     * A list of all [SelectedDatum] that are impacted by this event, sorted by distance.
     */
    public val selectedData: List<SelectedDatum<DOMAIN>>

    /**
     * An on-screen position coordinates in percentage, relative to the charting zone.
     */
    public val currentRelativePosition: RelativePoint

    /**
     * The zoom action (if any), can be determined depending on the user events from:
     *  - a dual-touch "pinch"
     *  - a mouse-wheel
     *  - a zoom selection on screen with mouse drag
     */
    public val zoomAction: ZoomAction

    /**
     * The pan action (if any), can be determined depending on the user events from:
     *  - a dual-touch movement
     *  - a mouse drag
     */
    public val panAction: PanAction

    /**
     * An on-screen zone rectangle coordinates representing the current dragging if any, relative to the charting zone.
     */
    public val dragZone: Rect?

    /**
     * An on-screen zone rectangle coordinates representing the current zooming if any, relative to the charting zone.
     * This zone is constrained by the current "zoomMode" constraints and display a more appropriate zone than dragZone.
     * This zone may contain "handlers" that are visual references to indicate how the zone can be extended.
     */
    public val zoomZone: RectWithHandlers?

    /**
     * A list of Datums without the marks information
     */
    public val data: List<Datum<DOMAIN>>

    /**
     * Stop the propagation of the base event that occurs on the platform.
     */
    public fun stopEventPropagation()

    /**
     * Force the propagation of the base event that occurs on the platform.
     */
    public fun forceEventPropagation()
}
