package io.data2viz.charts.logging

import io.data2viz.timeFormat.format
import kotlinx.datetime.Clock
import kotlin.jvm.JvmInline
import kotlin.time.ExperimentalTime
import kotlin.time.measureTime

/**
 * Information Logger.
 **/
@Suppress("unused")
@JvmInline
public value class DebugLogger(public val name: String) {

    /**
     * Print a message on the console. This message type is "verbose".
     */
    internal fun verbose(message: String) {
        println("$name ${timeInfo()}:: $message")
    }

    /**
     * Measure the time to process a [block] lambda and print the message with the elapsed time.
     */
    @OptIn(ExperimentalTime::class)
    public inline fun verbose(message: String, block: () -> Unit) {
        val executionTime = measureTime { block() }
        println("$name ${timeInfo()}:: $executionTime $message")
    }

    /**
     * Print a message on the console. This message type is "debug".
     */
    public fun debug(message: String) {
        println("$name ${timeInfo()}:: $message")
    }

    /**
     * Print a message on the console. This message type is "warn".
     */
    public fun warn(message: String) {
        println("$name ${timeInfo()}:: $message")
    }

    @PublishedApi
    internal companion object {
        private val formatter = format("%I:%M:%S.%L")
        private var lastLog = Clock.System.now().toEpochMilliseconds()

        private const val showDuration = false

        @PublishedApi
        internal fun timeInfo(): String = if (showDuration) {
            val now = Clock.System.now().toEpochMilliseconds()
            val diff = "${now - lastLog} ms".padStart(5, ' ')
            lastLog = now
            diff
        } else formatter(Clock.System.now())
    }
}
