package io.data2viz.charts.chart

import io.data2viz.charts.core.DirectedZoom
import io.data2viz.charts.core.Zoom
import io.data2viz.charts.core.computeZoom
import io.data2viz.scale.LinearScale
import io.data2viz.scale.Scales
import kotlin.test.Test
import kotlin.test.assertEquals
import kotlin.test.assertTrue

class TestZoomAndScale {

    private val epsilon = 1e-8

    @Test
    fun scaleAndZoom() {
        val domain = listOf(0.0, 10.0)
        val axisSize = listOf(.0, 100.0)

        val zoom = Zoom()

        val scale = Scales.Continuous.linear {
            this.domain = domain
            this.range = axisSize
        }

        assertEquals(20.0, scale(2.0))
    }

    @Test
    fun zoom_Identity_defaults() {
        var zoom = Zoom()
        assertEquals(1.0, zoom.xZoom.ratio)
        assertEquals(.0, zoom.xZoom.offset)

        zoom = Zoom.IDENTITY
        assertEquals(1.0, zoom.xZoom.ratio)
        assertEquals(.0, zoom.xZoom.offset)
    }

	@Test
	fun zoom_Identity_rescale() {
        val defaultScale = defaultScale()
        checkZoomAndInvertAndDomain(Zoom.IDENTITY.xZoom,            defaultScale,   defaultScale.domain)
    }

    @Test
    fun zoom_ratio_rescale() {
        val defaultScale = Scales.Continuous.linear {
            domain = listOf(.0, 100.0)
            range = listOf(.0, 100.0)
        }
        checkZoomAndInvertAndDomain(DirectedZoom(2.0, .0),          defaultScale, listOf(.0, 50.0))
        checkZoomAndInvertAndDomain(DirectedZoom(0.5, .0),          defaultScale, listOf(.0, 200.0))

        val invertedRangeScale = Scales.Continuous.linear {
            domain = listOf(.0, 100.0)
            range = listOf(1356.0, .0)
        }
        checkZoomAndInvertAndDomain(DirectedZoom(2.0,  .0),          invertedRangeScale, listOf(50.0, 100.0))
        checkZoomAndInvertAndDomain(DirectedZoom(0.5,  .0),          invertedRangeScale, listOf(-100.0, 100.0))
        checkZoomAndInvertAndDomain(DirectedZoom(10.0, .0),          invertedRangeScale, listOf(90.0, 100.0))
    }

    @Test
    fun zoom_offset_rescale() {
        checkZoomAndInvertAndDomain(DirectedZoom(1.0, 1.0),         defaultScale(), listOf(100.1, -99.9))
        checkZoomAndInvertAndDomain(DirectedZoom(1.0, 1000.0),      defaultScale(), listOf(200.0, .0))
        checkZoomAndInvertAndDomain(DirectedZoom(1.0, -1000.0),     defaultScale(), listOf(.0, -200.0))

        checkZoomAndInvertAndDomain(DirectedZoom(1.0, 1.0),         invertedScale(), listOf(100.1, -99.9).reversed())
        checkZoomAndInvertAndDomain(DirectedZoom(1.0, 1000.0),      invertedScale(), listOf(200.0, .0).reversed())
        checkZoomAndInvertAndDomain(DirectedZoom(1.0, -1000.0),     invertedScale(), listOf(.0, -200.0).reversed())
    }

    @Test
    fun zoom_offset_ratio_rescale() {
        checkZoomAndInvertAndDomain(DirectedZoom(2.0, 100.0),       defaultScale(), listOf(105.0, 5.0))
        checkZoomAndInvertAndDomain(DirectedZoom(2.0, -100.0),      defaultScale(), listOf(95.0, -5.0))
        checkZoomAndInvertAndDomain(DirectedZoom(10.0, .0),         defaultScale(), listOf(100.0, 80.0))
        checkZoomAndInvertAndDomain(DirectedZoom(10.0, 2000.0),     defaultScale(), listOf(120.0, 100.0))
        checkZoomAndInvertAndDomain(DirectedZoom(10.0, -2000.0),    defaultScale(), listOf(80.0, 60.0))

        checkZoomAndInvertAndDomain(DirectedZoom(2.0, 100.0),       invertedScale(), listOf(105.0, 5.0).reversed())
        checkZoomAndInvertAndDomain(DirectedZoom(2.0, -100.0),      invertedScale(), listOf(95.0, -5.0).reversed())
        checkZoomAndInvertAndDomain(DirectedZoom(10.0, .0),         invertedScale(), listOf(100.0, 80.0).reversed())
        checkZoomAndInvertAndDomain(DirectedZoom(10.0, 2000.0),     invertedScale(), listOf(120.0, 100.0).reversed())
        checkZoomAndInvertAndDomain(DirectedZoom(10.0, -2000.0),    invertedScale(), listOf(80.0, 60.0).reversed())
    }

    @Test
    fun zoom_specific_scales_rescale() {
        createScaleAndCheck(listOf(.0, 100.0),   listOf(.0, 100.0))
        createScaleAndCheck(listOf(100.0, .0),   listOf(.0, 100.0))
        createScaleAndCheck(listOf(-100.0, .0),  listOf(.0, 100.0))
        createScaleAndCheck(listOf(.0, -100.0),  listOf(.0, 100.0))

        createScaleAndCheck(listOf(.0, 100.0),   listOf(.0, -100.0))
        createScaleAndCheck(listOf(100.0, .0),   listOf(.0, -100.0))
        createScaleAndCheck(listOf(-100.0, .0),  listOf(.0, -100.0))
        createScaleAndCheck(listOf(.0, -100.0),  listOf(.0, -100.0))

        createScaleAndCheck(listOf(.0000001, .0),  listOf(.0, -100.0))
        createScaleAndCheck(listOf(-.0000001, .0), listOf(.0, -100.0))
        createScaleAndCheck(listOf(.0000001, .0),  listOf(.0, 100.0))
        createScaleAndCheck(listOf(-.0000001, .0), listOf(.0, 100.0))

        createScaleAndCheck(listOf(.0000001, .0),  listOf(.0, -100.0))
        createScaleAndCheck(listOf(-.0000001, .0), listOf(.0, -100.0))
        createScaleAndCheck(listOf(.0000001, .0),  listOf(.0, 100.0))
        createScaleAndCheck(listOf(-.0000001, .0), listOf(.0, 100.0))

        createScaleAndCheck(listOf(1000.0, .0),  listOf(1234.0, .0))
        createScaleAndCheck(listOf(-.00001, .0), listOf(1234.0, .0))
        createScaleAndCheck(listOf(.00001, .0),  listOf(-1234.0, .0))
        createScaleAndCheck(listOf(-.00001, .0), listOf(-1234.0, .0))

        createScaleAndCheck(listOf(1000.0, .0),  listOf(1234.0, -9876.0))
        createScaleAndCheck(listOf(-.00001, .0), listOf(1234.0, 9876.0))
        createScaleAndCheck(listOf(.00001, .0),  listOf(-1234.0, -9876.0))
        createScaleAndCheck(listOf(-.00001, .0), listOf(-1234.0, 9876.0))
    }

    /************************/

    /**
     * Defaultscale scale [100 .. -100] to [0 .. 2000]
     */
    private fun defaultScale(): LinearScale<Double> = Scales.Continuous.linear {
        domain = listOf(100.0, -100.0)
        range = listOf(.0, 2000.0)
    }

    /**
     * Inverted range scale (like a Y-AXIS) scale [100 .. -100] to [0 .. 2000]
     */
    private fun invertedScale(): LinearScale<Double> = Scales.Continuous.linear {
        domain = listOf(-100.0, 100.0)
        range = listOf(2000.0, .0)
    }

    private fun checkZoomAndInvertAndDomain(dZoom: DirectedZoom, scale: LinearScale<Double>, domain: List<Double>) {
        val newScale = dZoom.rescale(scale)
        assertEquals(domain, newScale.domain)
        assertEquals(scale.range, newScale.range)                       // range don't change

        val actualZoom = newScale.computeZoom(scale)
        assertTrue(actualZoom.ratio - epsilon < dZoom.ratio && dZoom.ratio < actualZoom.ratio + epsilon, "Ratio are not equals")
        assertTrue(actualZoom.offset - epsilon < dZoom.offset && dZoom.offset < actualZoom.offset + epsilon, "Offset are not equals")
    }

    private fun checkZoomAndInvert(dZoom: DirectedZoom, scale: LinearScale<Double>) {
        val newScale = dZoom.rescale(scale)
        assertEquals(scale.range, newScale.range)                       // range don't change

        val actualZoom = newScale.computeZoom(scale)
        assertTrue(actualZoom.ratio - epsilon < dZoom.ratio && dZoom.ratio < actualZoom.ratio + epsilon, "Ratio are not equals")
        assertTrue(actualZoom.offset - epsilon < dZoom.offset && dZoom.offset < actualZoom.offset + epsilon, "Offset are not equals")
    }

    private fun createScaleAndCheck(d: List<Double>, r: List<Double>) {
        checkMultipleZoomsOnScale(Scales.Continuous.linear {
            domain = d
            range = r
        })
    }

    private fun checkMultipleZoomsOnScale(scale: LinearScale<Double>) {
        checkZoomAndInvert(DirectedZoom(1.0, .0), scale)
        checkZoomAndInvert(DirectedZoom(100.0, .0), scale)
        checkZoomAndInvert(DirectedZoom(0.001, .0), scale)

        checkZoomAndInvert(DirectedZoom(1.0, 200.0), scale)
        checkZoomAndInvert(DirectedZoom(100.0, 200.0), scale)
        checkZoomAndInvert(DirectedZoom(0.001, 200.0), scale)

        checkZoomAndInvert(DirectedZoom(1.0, -200.0), scale)
        checkZoomAndInvert(DirectedZoom(100.0, -200.0), scale)
        checkZoomAndInvert(DirectedZoom(0.001, -200.0), scale)

        // Note the offset is not "0" but "-0" (which behave the same)
        checkZoomAndInvert(DirectedZoom(-1.0, .0), scale)
        checkZoomAndInvert(DirectedZoom(-2.0, .0), scale)
        checkZoomAndInvert(DirectedZoom(-.5, .0), scale)

        checkZoomAndInvert(DirectedZoom(-1.0, 200.0), scale)
        checkZoomAndInvert(DirectedZoom(-100.0, 200.0), scale)
        checkZoomAndInvert(DirectedZoom(-0.001, 200.0), scale)

        checkZoomAndInvert(DirectedZoom(-1.0, -200.0), scale)
        checkZoomAndInvert(DirectedZoom(-100.0, -200.0), scale)
        checkZoomAndInvert(DirectedZoom(-0.001, -200.0), scale)
    }

}
