package io.data2viz.charts.config

import io.data2viz.charts.config.internal.LocalOrDelegatedNullableProperty
import io.data2viz.charts.config.internal.LocalOrDelegatedProperty
import io.data2viz.charts.core.Font
import io.data2viz.charts.core.copyFont
import io.data2viz.color.Colors
import io.data2viz.math.pct
import io.data2viz.viz.FontFamily
import io.data2viz.viz.FontPosture
import io.data2viz.viz.FontWeight
import kotlin.test.Test
import kotlin.test.assertEquals
import kotlin.test.assertNotSame

class TestConfiguration {

    @Test
    fun testDefaultFont() {
        val font = Font()
        assertEquals(Colors.Web.black, font.fontColor)
        assertEquals(FontFamily.SANS_SERIF, font.fontFamily)
        assertEquals(FontPosture.NORMAL, font.fontStyle)
        assertEquals(FontWeight.NORMAL, font.fontWeight)
        assertEquals(12.0, font.fontSize)
    }

    @Test
    fun testDefaultConfiguration() {
        val conf = configuration()
        assertEquals(Font(), conf.copyFont())
    }

    @Test
    fun testTitleAsSameFontAsBase() {
        val testFont = FontFamily.specifiedFont("TestFont")

        val conf = configuration {
            fontFamily = testFont
        }
        assertEquals(testFont, conf.title.fontFamily)
        assertEquals(testFont, conf.legend.fontFamily)
        assertEquals(testFont, conf.mark.fontFamily)
        assertEquals(testFont, conf.tooltip.fontFamily)
        assertEquals(testFont, conf.x.fontFamily)
        assertEquals(testFont, conf.y.fontFamily)

        val georgia = FontFamily.specifiedFont("georgia")
        conf.fontFamily = georgia

        assertEquals(georgia, conf.title.fontFamily)
        assertEquals(georgia, conf.legend.fontFamily)
        assertEquals(georgia, conf.mark.fontFamily)
        assertEquals(georgia, conf.tooltip.fontFamily)
        assertEquals(georgia, conf.x.fontFamily)
        assertEquals(georgia, conf.y.fontFamily)


        conf.legend.fontFamily = testFont

        assertEquals(georgia, conf.title.fontFamily)
        assertEquals(testFont, conf.legend.fontFamily)
        assertEquals(georgia, conf.mark.fontFamily)
        assertEquals(georgia, conf.tooltip.fontFamily)
        assertEquals(georgia, conf.x.fontFamily)
        assertEquals(georgia, conf.y.fontFamily)


    }

    @Test
    fun testParentPropertyDelegate() {

        class ParentClass(var fontSize: Double = 12.0)

        class ChildClass(var parent: ParentClass) {
            var fontSize: Double by LocalOrDelegatedProperty({ parent }) { fontSize }
        }

        val parent1 = ParentClass(13.0)
        val parent2 = ParentClass(10.0)

        val child1 = ChildClass(parent1)
        assertEquals(13.0, child1.fontSize)
        child1.parent = parent2
        assertEquals(10.0, child1.fontSize)

        child1.fontSize = 11.0
        assertEquals(11.0, child1.fontSize)

        child1.parent = parent1
        assertEquals(11.0, child1.fontSize)

    }

    @Test
    fun testParentNullablePropertyDelegate() {

        class ParentClass(var strokeWidth: Double? = null)

        class ChildClass(val parent: ParentClass) {
            var strokeWidth: Double? by LocalOrDelegatedNullableProperty({ parent }) { parent.strokeWidth }
        }

        val parent1 = ParentClass(1.0)
        val child1 = ChildClass(parent1)
        assertEquals(1.0, child1.strokeWidth)

        child1.strokeWidth = null
        assertEquals(null, child1.strokeWidth)


        val parent2 = ParentClass(null)
        val child2 = ChildClass(parent2)
        assertEquals(null, child2.strokeWidth)
        child2.strokeWidth = 3.0
        assertEquals(3.0, child2.strokeWidth)
    }


    //Todo test overide color and stroke by title and legend

    @Test
    fun testFontHierarchy() {
        val config = configuration {
            fontSize = 13.0
        }
        assertEquals(13.0, config.fontSize)
        assertEquals(13.0 * 1.5, config.title.fontSize)
        assertEquals(13.0 * 1.2, config.legend.fontSize)
        assertEquals(13.0, config.tooltip.fontSize)
        assertEquals(13.0, config.x.fontSize)
        assertEquals(13.0, config.y.fontSize)
    }


    @Test
    fun testOverideFontHierarchy() {
        val config = configuration {
            fontSize = 13.0
            legend.fontSize = 10.0
            tooltip.fontSize = 10.0
            x.fontSize = 10.0
            y.fontSize = 10.0
        }
        assertEquals(13.0, config.fontSize)
        assertEquals(10.0, config.legend.fontSize)
        assertEquals(10.0, config.tooltip.fontSize)
        assertEquals(10.0, config.x.fontSize)
        assertEquals(10.0, config.y.fontSize)
    }

    @Test
    fun testStrokeHierarchy() {
        val config = configuration {
            strokeWidth = 5.0
        }
        assertEquals(5.0, config.strokeWidth)
        assertEquals(5.0, config.cursor.strokeWidth)
        assertEquals(5.0, config.mark.strokeWidth)
        assertEquals(5.0, config.x.strokeWidth)
        assertEquals(5.0, config.y.strokeWidth)
    }


    @Test
    fun testOverrideStrokeHierarchy() {
        val config = configuration {
            strokeWidth = 5.0
            cursor.strokeWidth = 6.0
            mark.strokeWidth = 6.0
            x.strokeWidth = 6.0
            y.strokeWidth = 6.0
        }

        assertEquals(5.0, config.strokeWidth)
        assertEquals(6.0, config.cursor.strokeWidth)
        assertEquals(6.0, config.mark.strokeWidth)
        assertEquals(6.0, config.x.strokeWidth)
        assertEquals(6.0, config.y.strokeWidth)
    }


    @Test
    fun testConfigurationCopy() {
        val config1: ChartConfig = configuration {
            fontColor = Colors.Web.red
        }

        val config2 = config1.deepCopy().apply {
        }

        config1.fontColor = Colors.Web.green

        assertEquals(Colors.Web.green, config1.x.fontColor)
        assertEquals(Colors.Web.green, config1.y.fontColor)

        assertEquals(Colors.Web.red, config2.fontColor)
        assertEquals(Colors.Web.red, config2.x.fontColor)
        assertEquals(Colors.Web.red, config2.y.fontColor)

    }


    @Test
    fun testConfigurationComposition() {
        val config: ChartConfig = configuration {

            fontSize = 15.0

            xAxis {
                fontSize = 10.0
                bandwidthRatio = 20.pct
            }
        }

        val cursor = config.cursor
        val events = config.events
        val legend = config.legend
        val mark = config.mark
        val selection = config.selection
        val tooltip = config.tooltip
        val x = config.x
        val y = config.y
        val zoom = config.zoom

        val copy = config.deepCopy()
        assertNotSame(copy.cursor, cursor)
        assertNotSame(copy.events, events)
        assertNotSame(copy.legend, legend)

        assertNotSame(copy.mark, mark)
        assertNotSame(copy.selection, selection)
        assertNotSame(copy.tooltip, tooltip)
        assertNotSame(copy.x, x)
        assertNotSame(copy.y, y)
        assertNotSame(copy.zoom, zoom)

        assertEquals(copy.cursor, cursor)
        assertEquals(copy.events, events)
        assertEquals(copy.legend, legend)
        assertEquals(copy.mark, mark)
        assertEquals(copy.selection, selection)
        assertEquals(copy.tooltip, tooltip)
        assertEquals(copy.zoom, zoom)
        assertEquals(copy.x, x)
        assertEquals(copy.y, y)


    }


    @Test
    fun modifyDefaultFont() {
        val conf = configuration {
            fontColor = Colors.Web.red
        }

        assertEquals(Colors.Web.red, conf.fontColor)
    }


}
